# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

__all__ = ['Schema', 'getschema']

"""Base schema package."""

from six import add_metaclass, string_types

from b3j0f.conf import Configurable

_SCHEMACLS = []


class MetaSchema(type):
    """Automatically register schemas."""

    def __new__(mcs, name, bases, _dict):

        result = super(MetaSchema, mcs).__new__(mcs, name, bases, _dict)

        if bases != (dict,):
            _SCHEMACLS.append(result)

        return result

@Configurable(paths='schema.conf')
@add_metaclass(MetaSchema)
class Schema(dict):
    """Base class for schema.

    Couple of (key, value) are (Property name and value).
    Referenced schema are instance of Schema.

    other attributes are:

    - name: schema name.
    - uid: schema unique id.
    - resource: schema resource from where it has been loaded.
    - _schema : specific schema object."""

    def __init__(self, resource, properties=None, *args, **kwargs):
        """
        :param resource: resource from where get schema content.
        :param Properties properties: list of Property.
        """

        super(Schema, self).__init__(*args, **kwargs)

        self.resource = resource

        if properties is not None:
            for prop in properties:
                self[prop.name] = prop

    def validate(self, data):
        """Validate input data.

        :param dict data: data to validate with this schema."""

        result = True

        names = set(data)

        for name in self:
            prop = self[name]

            if name not in data and prop.mandatory:
                result = False

            else:
                value = data[name]
                result = prop.validate(value)

            names.remove(name)

            if not result:
                break

        else:
            result = not names

        return result

    def newdata(self, **properties):
        """Instanciate a new data with input properties."""

        result = properties.copy()

        for name in self:

            if name not in properties:

                prop = self[name]

                if prop.mandatory:
                    result[name] = prop

        return result

    def save(self, resource=None):
        """Save this schema in the input resource.

        :param resource: default is this resource.
        """

        if resource is None:
            resource = self.resource

        self._save(resource=resource)

    def _save(self, resource):
        """Method to override in order to save this schema in the input resource
        media.

        :param resource: resource to write this schema.
        """

        raise NotImplementedError()


def getschema(resource, *args, **kwargs):
    """Schema factory.

    Find the right schema class to load input resource.

    :param resource: resource from where load schema.
    :param args: schema args.
    :param kwargs: schema kwargs"""

    for schemacls in _SCHEMACLS:

        try:
            result = schemacls(resource=resource, *args, **kwargs)

        except Exception:
            continue

        break

    else:
        raise ValueError('No parser found for {0}.'.format(resource))

    return result



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Schema property package."""

from numbers import Number

from six import string_types

from collections import Iterable


class Property(object):
    """Schema property"""

    def __init__(
            self, name, ptype=object, default=None, mandatory=True,
            *args, **kwargs
    ):
        """
        :param str name: property name.
        :param type ptype: property type.
        :param default: default value.
        :param bool mandatory: mandatory property characteristic.
        """

        super(Property, self).__init__(*args, **kwargs)

        self.name = name
        self.ptype = ptype
        self.default = default
        self.mandatory = mandatory

    def validate(self, data):
        """Validate input data."""

        return isinstance(data, self.ptype)


class ArrayProperty(Property):
    """Array property type"""

    PTYPE = 'array'

    def __init__(
            self, cardinality=None, itemtype=None, ptype=Iterable,
            *args, **kwargs
    ):
        """
        :param int(s) cardinality: (min and) max number of items.
        :param type itemtype: type of items.

        """

        super(ArrayProperty, self).__init__(ptype=ptype, *args, **kwargs)

        self.cardinality = cardinality
        self.itemtype = itemtype
        self.ptype = ptype

    def validate(self, data, *args, **kwargs):

        result = super(ArrayProperty, self).validate(data=data, *args, **kwargs)

        if result:
            result = not isinstance(data, string_types)

            if result:
                if self.cardinality is not None:
                    if isinstance(self.cardinality, Number):
                        result = len(data) <= self.cardinality

                    else:
                        result = (
                            self.cardinality[0] <= len(data) <=
                            self.cardinality[1]
                        )

                if result and self.itemtype is not None:

                    itemtype = self.itemtype

                    for item in data:
                        result = isinstance(item, itemtype)

                        if not result:
                            break

        return result


class FunctionProperty(Property):
    """Function property type."""

    def __init__(self, args=None, rtype=None, *vargs, **kwargs):
        """
        :param list args: argument names.
        :param type rtype: function result type.
        """

        super(FunctionProperty, self).__init__(*vargs, **kwargs)

        self.args = args
        self.rtype = rtype


class SchemaProperty(Property):
    """Schema property."""

    def __init__(self, schema, *args, **kwargs):
        """
        :param type schema: schema type.
        """

        super(SchemaProperty, self).__init__(*args, **kwargs)

        self.schema = schema

    def validate(self, data):

        result = super(SchemaProperty, self).validate(data)

        if result:
            result = self.schema.validate(data)

        return result



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from __future__ import absolute_import

"""json schema module."""

from ..base import Schema
from ..property import Property

from json import loads, load, dump

from jsonschema import validate


class JSONProperty(Property):

    def __init__(self, jsonproperty, *args, **kwargs):

        super(JSONProperty, self).__init__(*args, **kwargs)

        self.jsonproperty = jsonproperty


class JSONSchema(Schema):
    """Schema for json resources."""

    def __init__(self, *args, **kwargs):

        super(JSONSchema, self).__init__(*args, **kwargs)

        try:
            self._schema = loads(self.resource)

        except TypeError:
            with open(self.resource) as jsonfile:
                self._schema = load(jsonfile)

        self.uid = self._schema['id']

        for name in self._schema['property']:

            jsonproperty = self._schema['property'][name]

            self[name] = JSONProperty(name=name, jsonproperty=jsonproperty)

    def validate(self, data):

        return validate(data, self._schema)

    def save(self, resource):

        with open(resource, "w") as rstream:
            dump(self._schema, rstream)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Class schema package."""

from inspect import getargspec, isroutine, isclass

from b3j0f.utils.path import getpath

from .base import Schema
from .property import FunctionProperty, Property, SchemaProperty


class PythonFunctionProperty(FunctionProperty):
    """Python function property class."""

    def __init__(self, func, *args, **kwargs):

        super(PythonFunctionProperty, self).__init__(*args, **kwargs)

        self.func = func

        try:
            self.args, self.vargs, self.kwargs, self.default = getargspec(func)

        except TypeError:
            self.args, self.vargs, self.kwargs, self.default = (), (), {}, ()

    def __call__(self, instance=None, args=None, kwargs=None):
        """Execute this function property.

        :param instance: func instance if func is a method.
        :param tuple args: func args.
        :param dict kwargs: func kwargs."""

        if args is None:
            args = ()

        if kwargs is None:
            kwargs = {}

        if instance is None:
            result = self.func(*args, **kwargs)

        else:
            result = self.func(instance, *args, **kwargs)

        return result


class ClassSchema(Schema):
    """Class schema."""

    def __init__(self, public=True, *args, **kwargs):
        """
        :param bool public: if True (default), convert only public attributes.
        """

        super(ClassSchema, self).__init__(*args, **kwargs)

        if not isclass(self.resource):
            raise TypeError('resource {0} is not a class'.format(self.resource))

        self.uid = getpath(self.resource)
        self.name = self.resource.__name__
        self.public = public

        for name in dir(self.resource):

            if self.public and name.startswith('_'):
                continue

            attribute = getattr(self.resource, name)

            if isinstance(attribute, Schema):
                prop = SchemaProperty(name=name, schema=attribute)

            elif isroutine(attribute):
                prop = PythonFunctionProperty(name=name, func=attribute)

            else:
                prop = Property(name=name, ptype=type(attribute))

            self[name] = prop

            if name == '__slots__':
                for name in self.resource.__slots__:
                    prop = Property(name=name)
                    self[name] = prop

        self._schema = self

    def validate(self, data):

        return isinstance(data, self.resource)

    def newdata(self, **properties):

        return self.resource(**properties)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------


from unittest import main

from b3j0f.utils.ut import UTCase

from ..base import getschema
from ..property import Property, SchemaProperty, FunctionProperty
from ..cls import ClassSchema, PythonFunctionProperty


class ClassSchemaTest(UTCase):

    def test(self):

        class Test(object):

            a = 'a'

            def b(self, b):
                return b

        test = Test()

        schema = getschema(Test)

        self.assertIsInstance(schema, ClassSchema)

        self.assertIsInstance(schema['a'], Property)
        self.assertNotIsInstance(schema['a'], FunctionProperty)
        self.assertNotIsInstance(schema['a'], PythonFunctionProperty)

        self.assertIsInstance(schema['b'], Property)
        self.assertIsInstance(schema['b'], FunctionProperty)
        self.assertIsInstance(schema['b'], PythonFunctionProperty)

        self.assertEqual(schema['b'](test, [2]), 2)
        self.assertEqual(schema['b'](test, None, {'b': 2}), 2)

        self.assertEqual(len(schema), 2)

        self.assertNotIn('c', schema)

        Test.c = schema

        schema = getschema(Test)

        self.assertIsInstance(schema, ClassSchema)

        self.assertIsInstance(schema['a'], Property)
        self.assertNotIsInstance(schema['a'], FunctionProperty)
        self.assertNotIsInstance(schema['a'], PythonFunctionProperty)

        self.assertIsInstance(schema['b'], Property)
        self.assertIsInstance(schema['b'], FunctionProperty)
        self.assertIsInstance(schema['b'], PythonFunctionProperty)

        self.assertIsInstance(schema['c'], Property)
        self.assertIsInstance(schema['c'], SchemaProperty)

        self.assertEqual(len(schema), 3)

        schema = getschema(Test, public=False)

        self.assertEqual(len(schema), len(dir(Test)))

if __name__ == '__main__':
    main()
